let debug = false
let panic_mode = ref false
let panic_at   = 20 (* seconds before limit *)
let panic_exit =  7 (* seconds before limit *)


(* The character with its property record *)
type decorated = { 
    char : char; 
    b : bool;
    i : bool;
    em : bool;
    s : bool; 
    tt : bool;
    u : int; (* 0 to 3 *)
    size : int; (* -1 to 9 *)
    color : char; (* -rgbcmykw *)
  } 

(* The color and size values at the start *)
let default_color = '-'
let default_size  = -1

(* The starting or root context *)
let rootContext = {
  char = '.';
  b = false;
  em = false;
  i = false;
  s = false;
  tt = false;
  u = 0;
  size = default_size;
  color = default_color;
}

let isSpace c = c <= ' '

(* For debugging purposes *)
let print_decorated_char dc =
  if isSpace dc.char && not dc.tt then prerr_char ' ' else prerr_char dc.char;
  prerr_string " -";
  prerr_string (if dc.b then  " b" else "  "); 
  prerr_string (if dc.em then " em" else "   ");
  prerr_string (if dc.i then " i" else "  ");
  prerr_string (if dc.s then " s" else "  ");
  prerr_string (if dc.tt then " tt" else "   ");
  if dc.u > 0 
  then begin prerr_string " u"; prerr_int dc.u; end
  else prerr_string "   ";
  if dc.size >= 0 
  then begin prerr_string " "; prerr_int dc.size end
  else prerr_string "  ";
  if dc.color != '-' 
  then begin prerr_string " "; prerr_char dc.color; end
  else prerr_string "  ";
  prerr_char '\n';
  ()

let assertprint s =
  flush stdout;
  prerr_string s; 
  prerr_char '\n';
  flush stderr;
  false



(* Context constructors *)
let c_bold context = {context with b=true}
let c_em context   = {context with em=if not context.s then not context.em else context.em}
let c_s  context   = {context with s=true}
let c_i  context   = {context with i=true}
let c_tt context   = {context with tt=true}
let c_u  context   = {context with u=min 3 (succ context.u)}
let c_size context n = 
  (assert (n = default_size || n >= 0 && n <= 9));
  {context with size=n}
let c_color context c = 
  (assert 
     ((c = default_color) ||
     (match c with 'r'|'g'|'b'|'c'|'m'|'y'|'k'|'w' -> true 
     | _ -> Printf.eprintf "Bad color: %c\n" c; false)));
  {context with color=c}
let c_pl context   = {rootContext with color=context.color; size=context.size}


(* The formal way of comparing redundant spaces *)
let isSpaceEqual_formal s1 s2 =
  let spaceContext c =
    {c with 
      char = ' ';
      s=false; em=false; b=false; i=false; 
      color = if c.u <= 0 then 'w' else c.color;}
  in
  isSpace s1.char && isSpace s2.char&&
  (spaceContext s1) = (spaceContext s2)

(* ... my way *)
let isSpaceEqual s1 s2 =
  let ret =
    isSpace s1.char && isSpace s2.char &&
    s1.tt = s2.tt && s1.size = s2.size &&
    s1.u = s2.u && (s1.u = 0 || s1.color = s2.color)
  in
  assert ((ret = isSpaceEqual_formal s1 s2) || begin print_decorated_char s1; print_decorated_char s2; assertprint "isSpaceEqual_formal" end);
  ret

(* A char with current context *)
let decorate context c = 
  {context with char=c}





(* Read characters until end-of-tag *)
let read_tag () =
  let c = input_char stdin in
  while (input_char stdin) != '>' do () done;
  c

(* Heh *)
let between lo x hi = lo <= x && x <= hi

(* Read stdin and return a list of rendered characters *)
let read_sml () : decorated list =
  let eofChar = '\004' in
  let rec read_loop context c_stack acc =
    match (try input_char stdin with _ -> eofChar) with
    | c when c = eofChar -> 
	assert (context = rootContext || assertprint "Unclosed tags");
	List.rev acc
    | '<' -> begin
      match  read_tag () with
      | 'B' -> read_loop (c_bold context) (context :: c_stack) acc
      | 'E'-> read_loop (c_em context) (context :: c_stack) acc
      | 'I' -> read_loop (c_i context) (context :: c_stack) acc
      | 'P'-> read_loop 
	    (c_size (c_color rootContext context.color) context.size) 
	    (context :: c_stack) acc
      | 'S' -> read_loop (c_s context) (context :: c_stack) acc
      | 'T'-> read_loop (c_tt context) (context :: c_stack) acc
      |	'U' -> read_loop (c_u context) (context :: c_stack) acc
      |	n when between '0' n '9' -> read_loop 
	    (c_size context ((int_of_char n) - (int_of_char '0')))
	    (context :: c_stack) acc
      |	'/' -> 
	  (match c_stack with h::t -> read_loop h t acc | [] -> failwith "Too many close-tags")
      |	c -> read_loop (c_color context c) (context :: c_stack) acc
    end
    | space when space <= ' ' ->
	let prev = match acc with h :: t -> h | [] -> context in
	let dc = {prev with 
		   char = space;
		   tt = context.tt;
		   size = context.size;
		   u = context.u} in
	let newacc =
	  match acc with 
	  | sp :: t when (not context.tt) && (isSpaceEqual sp dc) -> acc
	  | _ -> dc :: acc
	in read_loop context c_stack newacc
    | c -> read_loop context c_stack ((decorate context c) :: acc)	  
  in
  read_loop rootContext [] []


let attr_b = 1
let attr_i = 2
let attr_em = 3
let attr_s = 4
let attr_tt = 5
let attr_u = 6
let attr_size = 7
let attr_color = 8
let attr_pl = 9

let attr_tag = [| "<dummy>"; "<B>"; "<I>"; "<EM>"; "<S>"; "<TT>"; "<U>"; "<size?>"; "<clr?>"; "<PL>" |]
let untag s = 
  let len = String.length s in
  let str = String.make (len+1) '<' in
  String.blit s 1 str 2 (len - 1);
  str.[1] <- '/';
  str
let mktag s =
  let len = String.length s in
  let str = String.make (len+2) '<' in
  String.blit s 0 str 1 len;
  str.[len+1] <- '>';
  str
let mktagc c =
  let len = 1 in
  let str = String.make (len+2) '<' in
  str.[1] <- c;
  str.[len+1] <- '>';
  str

let attr_untag = Array.map untag attr_tag

let length_of_untags c1 c2 =
  let iob b y n = if b then y else n in
  (iob (c1.b > c2.b) 7 0) +
    (iob (c1.i > c2.i) 7 0) +
    (iob (c1.u > c2.u) ((c1.u - c2.u) * 7) 0) +
    (iob (c1.s > c2.s) 7 0) +
    (iob (c1.tt > c2.tt) 9 0) +
    (iob (c1.em > c2.em) 0 0) + (* !!! *)
    0

(******************************************************************************)
(* Da optimizer *)

let different_branch' (context:decorated) (c:decorated) (context_stack:decorated list) =
  context.b > c.b ||
  context.i > c.i ||

  context.em > c.em || (* !! *)

  context.s > c.s ||
  context.tt > c.tt ||
  context.u > c.u ||
  (context.size != c.size && (List.exists (fun cc -> cc.size = c.size) context_stack)) || 
  (context.color != c.color && (List.exists (fun cc -> cc.color = c.color) context_stack)) ||  
  false

let same_branch con c context_stack =
  (con.em != c.em &&
   con.b = c.b &&
   con.i = c.i &&
   con.u = c.u &&
   con.s = c.s &&
   con.tt = c.tt &&
   (con.size = c.size || (not (List.exists (fun cc -> cc.size = c.size) context_stack))) &&
   (con.color = c.color || (not (List.exists (fun cc -> cc.color = c.color) context_stack)))
  ) ||
  (con.b <= c.b &&
   con.i <= c.i &&
   con.u <= c.u &&
   con.s <= c.s &&
   con.em <= c.em &&
   con.tt <= c.tt &&
   (con.size = c.size || (not (List.exists (fun cc -> cc.size = c.size) context_stack))) &&
   (con.color = c.color || (not (List.exists (fun cc -> cc.color = c.color) context_stack))))  

let different_branch (context:decorated) (c:decorated) (context_stack:decorated list) =
  not (same_branch context c context_stack)
(*
  (not (context.b = c.b && context.i = c.i && context.s = c.s && context.tt = c.tt &&
	context.u = c.u && context.color = c.color && context.size = c.size && 
	context.em != c.em ))  
  &&

  (context.b > c.b ||
  context.i > c.i ||
  context.s > c.s ||
  context.tt > c.tt ||
  context.u > c.u ||
  (context.size != c.size && (List.exists (fun cc -> cc.size = c.size) context_stack)) || 
  (context.color != c.color && (List.exists (fun cc -> cc.color = c.color) context_stack)) ||  
  false)
*)

let rec exists_in_branch (f : decorated -> bool) context context_stack doc =
  match doc with
  | [] -> false
  | h :: t when different_branch context h context_stack -> false
  | h :: t -> (f h) || (exists_in_branch f context context_stack t)

and  decide_pl context context_stack doc =
  let ret =
    exists_in_branch (fun c -> {c with char=context.char} = context  && not (isSpace c.char)) (c_pl context) context_stack doc
  in
  (* Printf.eprintf "decide_pl = %b\n" ret; *)
  ret

let get_duration (context:decorated) (doc:decorated list) (context_stack:decorated list) =
  let rec dur_loop con doc acc =
    match doc with
    | [] -> acc
    | h :: t 
      when different_branch context h context_stack
	  && not (length_of_untags con h > 9 && decide_pl context context_stack doc) 
      -> acc
    | h :: t -> dur_loop h t (acc + 1)
  in dur_loop context doc 0



let print_sml doc =
  let rec print_loop 
      (context:decorated) (context_stack:decorated list) 
      (close_stack:string list) 
      (doc:decorated list) : unit =
    let rec add_tag dc rest =
      let needtag = ref 0 in
      let newcontext = ref context in
      let needtagduration = ref ~-1 in
      let testDuration c num =
	let d = get_duration c doc context_stack in
	if d > !needtagduration then begin 
	  needtag := num; newcontext := c; needtagduration := d
	end
      in
      match close_stack with
      |	"</PL>" :: _ 
	when 
	  not (isSpace dc.char) &&	  
	  {(List.hd context_stack) with char = dc.char} = dc
	  -> remove_tag ()	  
      |	_ -> begin      
	  if (dc.b > context.b)  then testDuration (c_bold context) 1;	  
	  if (dc.i > context.i)  then testDuration (c_i context) 2;	  
	  if (dc.s > context.s)  then testDuration (c_s context) 4;	  
	  if (dc.tt> context.tt) then testDuration (c_tt context) 5;	  
	  if (dc.u > context.u)  then testDuration (c_u context) 6;
	  if (dc.em != context.em) then testDuration (c_em context) 3;	  

	  if !needtag > 0 then begin
	    print_string attr_tag.(!needtag);
	    print_loop !newcontext (context :: context_stack) 
	      (attr_untag.(!needtag) :: close_stack) doc
	  end
	  else begin
	    assert ((decorate context dc.char) = dc || 
	    (isSpaceEqual (decorate context dc.char) dc) ||
	    begin 
	      print_char '\n';
	      print_decorated_char dc;
	      print_decorated_char context;	      
	      assertprint "decorated char is not decorated char";
	    end);
	    print_char dc.char;
	    if debug && false then begin
	      print_decorated_char context;
	      Printf.eprintf "\t\t%c => %d\n" dc.char (get_duration context doc context_stack);
	    end;
	    print_loop context context_stack close_stack rest
	  end
      end

    and remove_tag () =
      match context_stack, close_stack with
      | con :: t, str :: t2 -> print_string str; print_loop con t t2 doc
      | _ -> failwith "print_loop can't pop"

    in

    match doc with
    | [] -> List.iter print_string close_stack
    | dc :: rest -> begin	
	(* If any removed attribute, pop stack and recurse with doc *)
	(* Toggle <EM> instead of removing it if its used again on the same branch *)
	(* For a new color, If its on stack, pop and recurse *)
	(* If all tags (more than 1 open) gone, and they reappear together, use <PL> *)

	let len_un = length_of_untags context dc in 
	if 
	  len_un > 0 || 
(*
	  dc.b < context.b || dc.i < context.i ||
          dc.s < context.s || dc.tt < context.tt || 
          dc.u < context.u ||
*)
	  (* dc.em < context.em || *)
	  (dc.size  != context.size  && (List.exists (fun c -> c.size = dc.size) context_stack)) || 
	  (dc.color != context.color && (List.exists (fun c -> c.color=dc.color) context_stack)) ||
	  false
	then 
	  if len_un > 9 && decide_pl context context_stack doc then begin
	    print_string attr_tag.(attr_pl);
	    print_loop (c_pl context) (context :: context_stack) 
	      (attr_untag.(attr_pl) :: close_stack) doc
	  end
	  else 
	    remove_tag ()	      
	else if dc.em < context.em then
	  match close_stack with
	  | "</EM>" :: _ -> 
	      (* add_tag ~avoidEM:true dc rest*)
	      remove_tag () 
	  | _ -> add_tag dc rest

	else if  dc.em != context.em then begin
	  if 
	    (exists_in_branch (fun c -> {c with char=context.char} = context) dc (context :: context_stack)  doc) ||  
	    not (List.exists (fun c -> c.em = dc.em) context_stack)  (* if popping won't help *)
	  then add_tag dc rest
	  else remove_tag ()
	end
	else if dc.size != context.size then begin
	  assert (dc.size >= 0);
	  let tag  = mktagc (char_of_int ((int_of_char '0') + dc.size)) in
	  print_string tag;
	  print_loop (c_size context dc.size) (context :: context_stack) ((untag tag) :: close_stack) doc
	end
	else if dc.color != context.color then begin
	  assert (dc.color != '-');
	  let tag  = mktagc dc.color in
	  print_string tag;
	  print_loop (c_color context dc.color) (context :: context_stack) ((untag tag) :: close_stack) doc
	end
	else add_tag dc rest
    end;

	(* What about <S> ... arrgh *)
	(* For each added attribute, choose longest and recurse with doc *)
	
  in
  print_loop rootContext [] [] doc;
  ()
(******************************************************************************)

(* Main main main *)

let alarm2 = Sys.Signal_handle (fun _ -> exit 2)
let alarm1 = Sys.Signal_handle 
    (fun _ -> 
      panic_mode := true; 
      Sys.set_signal Sys.sigalrm alarm2;
      ignore (Unix.alarm (panic_at - panic_exit)))

let main () =
  let tot_time =  (try int_of_string (Sys.getenv "TLIMIT") with _ -> 0) in
  if tot_time > panic_at then ignore (Unix.alarm (tot_time - panic_at));
  Sys.set_signal Sys.sigalrm alarm1;  
  
  let doc = read_sml () in
  if debug then begin prerr_string "Done parsing\n"; flush stderr end;
  if (Array.length Sys.argv) <= 1 || (Sys.argv.(1) <> "-c")
  then print_sml doc
  else List.iter print_decorated_char doc;
  ()

(* Go go go *)
;; main () ;;
